from lxml import html
import requests
import urllib.parse
import weblogin
import weblogin.seamlessaccess as sa

class UGlogin(weblogin.AutologinHandler):
  """
  Login handler (weblogin.AutologinHandler) for UG logins, i.e. through 
  login.ug.kth.se.
  """
  LOGIN_URL = "https://login.ug.kth.se"
  
  def __init__(self, username, password, login_trigger_url=None,
               rerun_requests=False):
    """
    Creates a login handler that automatically logs into KTH.
    - Requires username and password.
    - Optional `login_trigger_url` is a page that redirects to the login page,
      for instance, the API URLs don't redirect, but the UI URLs do.
    - Optional `rerun_requests` specifies whether we want to rerun the original 
      request that triggered authentication failure.
    """
    super().__init__()
    self.__username = username
    self.__password = password
    self.__login_trigger_url = login_trigger_url
    self.__rerun_requests = rerun_requests
    self.__logging_in = False

  def need_login(self, response):
    """
    Checks a response to determine if logging in is needed,
    returns True if needed
    """
    if self.__logging_in:
      return False

    if response.status_code == requests.codes.unauthorized and \
              "kth.se" in response.url:
      self.__rerun_requests = True
      return True
    elif response.url.find(self.LOGIN_URL) == 0:
      return True

    return False

  def login(self, session, response, args=[], kwargs={}):
    """
    Performs a login based on the response `response` from a request to session 
    `session`.
    `args` and `kwargs` are the options from the request triggering the login 
    procedure, this is so that we can redo that request after logging in.

    Raises an AuthenticationError exception if authentication fails.
    """
    self.__logging_in = True
    if response.status_code == requests.codes.unauthorized:
      trigger_response = session.get(self.__login_trigger_url)
      login_response = self.login(session, trigger_response)
    else:
      doc_tree = html.fromstring(response.text)
      login_form = doc_tree.xpath("//form[@id='loginForm']")
      if len(login_form) < 1:
        form = doc_tree.xpath("//form")[0]

        data = {}
        for variable in form.xpath("//input"):
          if variable.name:
            data[variable.name] = variable.value or ""

        action_url = urllib.parse.urljoin(response.url, form.action)

        saml_response = session.request(form.method, action_url, data=data)

        if saml_response.status_code != requests.codes.ok:
          raise weblogin.AuthenticationError(
                              f"SAML error: not OK response: {saml_response.text}")

        final_response = saml_response
      else:
        login_form = login_form[0]
        data = {}

        for variable in login_form.xpath("//input"):
          if variable.value:
            data[variable.name] = variable.value

        data["UserName"] = self.__username if "@ug.kth.se" in self.__username \
                                           else self.__username + "@ug.kth.se"
        data["Password"] = self.__password
        data["Kmsi"] = True

        login_response = session.request(
          login_form.method, f"{self.LOGIN_URL}/{login_form.action}",
          data=data)

        if login_response.status_code != requests.codes.ok:
          raise weblogin.AuthenticationError(
            f"authentication as {self.__username} to {login_response.url} failed: "
            f"{login_response.text}")

        login_doc_tree = html.fromstring(login_response.text)
        login_form = login_doc_tree.xpath("//form[@id='loginForm']")
        if len(login_form) > 0:
          raise weblogin.AuthenticationError(
            f"authentication as {self.__username} failed (redirect to same page), "
            f"probably wrong username or password.")

        final_response = login_response
    self.__logging_in = False

    if self.__rerun_requests and args:
      return session.request(*args, **kwargs)
    return final_response
class SAMLlogin(weblogin.AutologinHandler):
  """
  Login handler (weblogin.AutologinHandler) for SAML at KTH. This will relay to 
  UG (login.ug.kth.se) which handles the password-based authentication.
  """

  def __init__(self, rerun_requests=False):
    """
    Creates a login handler that automatically handles the SAML requests used 
    at KTH.
    """
    super().__init__()
    self.__rerun_requests = rerun_requests
    self.__logging_in = False

  def need_login(self, response, ignore_logging_in=False):
    """
    Checks a response to determine if we should handle a request.
    Returns True if needed.
    """
    if self.__logging_in:
      return False

    return "saml" in response.url and "sys.kth.se" in response.url

  def login(self, session, response, args=[], kwargs={}):
    """
    - Performs an action based on the response `response` from a request to 
      session `session`.
    - `args` and `kwargs` are the options from the request triggering the login 
      procedure, this is so that we can redo that request after logging in.
    - Raises an AuthenticationError exception on fails.
    """
    self.__logging_in = True
    doc_tree = html.fromstring(response.text)
    try:
      form = doc_tree.xpath("//form")[0]
    except IndexError:
      pass
    else:
      data = {}
      for variable in form.xpath("//input"):
        if variable.name:
          data[variable.name] = variable.value or ""

      action_url = urllib.parse.urljoin(response.url, form.action)
      saml_response = session.request(form.method, action_url, data=data)

      if saml_response.status_code != requests.codes.ok:
        raise weblogin.AuthenticationError(
                            f"SAML error: not OK response: {saml_response.text}")
    self.__logging_in = False

    if self.__rerun_requests and args:
      return session.request(*args, **kwargs)
    return saml_response



import requests

def get_entity_data_by_id(id):
  """
  Requests entity data from SeamlessAccess.org for entity with unique ID `id`.
  Returns JSON (dictionary) containing data.
  """
  response = requests.get(f"https://md.seamlessaccess.org/entities/{id}.json")
  try:
    return response.json()
  except:
    raise Exception(f"invalid response from SeamlessAccess.org for ID {id}")
def find_entity_data_by_name(name):
  """
  Searches SeamlessAccess.org for an institution by name `name`.
  Returns a list of institutions' data.
  """
  response = requests.get(f"https://md.seamlessaccess.org/entities/?q={name}")
  try:
    return response.json()
  except:
    raise Exception(f"invalid response from SeamlessAccess.org for name {name}")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['weblogin']

package_data = \
{'': ['*']}

install_requires = \
['lxml>=4.9.1,<5.0.0', 'requests>=2.28.1,<3.0.0']

setup_kwargs = {
    'name': 'weblogin',
    'version': '1.4',
    'description': 'Automates logging into web UIs to access unofficial APIs',
    'long_description': 'We want to use APIs from web UIs that require login. What we want to do\nis to set up a session, login and then use the API of the page. The\nreason we want to do this is to use existing APIs. For instance, the\nuser group management service at KTH, see\n[1](#UGEditor){reference-type="ref" reference="UGEditor"}. We can then\nuse the service, track the requests in the browser\'s developer tools.\nThen we can simply make the same requests from Python.\n\n![Screenshot of the KTH UG Editor with Firefox\'s Developer Tools open,\nshowing network requests made.](https://github.com/dbosk/weblogin/raw/main/doc/figs/ug.png){#UGEditor width="\\\\columnwidth"}\n\nFor instance, we can redo the request in\n[1](#UGEditor){reference-type="ref" reference="UGEditor"} like this:\n\n``` {.python}\nfrom weblogin.kth import AutologinSession\nimport os\n\nug = AutologinSession(os.environ["KTH_LOGIN"],\n                      os.environ["KTH_PASSWD"],\n                      "https://app.kth.se/ug-gruppeditor/")\n\nresponse = ug.get("https://app.kth.se/ug-gruppeditor/api/ug/users"\n                  "?filter=memberOf eq \'u26yk1i3\'")\n```\n\nThe code above will access the API used by the KTH UG group editor\nservice. It will automatically sign in when needed. The API URLs don\'t\ntrigger a redirect to log in, they just give a 401 unauthorized error.\nHowever, we can use the main URL to the UI to trigger such an event, log\nin and then access the API. All this happens automatically in the\nbackground.\n\nThe way we do this is to subclass the `requests.Session` class to\nintercept all requests of a session to check for signs indicating that\nwe must log in. When we detect such sign, we log in and resume as if\nnothing ever happened.\n',
    'author': 'Daniel Bosk',
    'author_email': 'dbosk@kth.se',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dbosk/weblogin',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
